@@ -1,56 +1,21 @@
-use crate::no_std::ToOwned;
-use core::borrow::Borrow;
 use core::fmt;
 
+use crate::de::{NumberVisitor, TypeHint, ValueVisitor};
 use crate::error::Error;
-use crate::expecting::{self, BadVisitorType, Expecting, InvalidType};
+use crate::expecting::{self, Expecting};
+use crate::mode::Mode;
 
-/// A visitor for data where it might be possible to borrow it without copying
-/// from the underlying [Decoder].
-///
-/// A visitor is required with [Decoder::decode_bytes] and
-/// [Decoder::decode_string] because the caller doesn't know if the encoding
-/// format is capable of producing references to the underlying data directly or
-/// if it needs to be processed.
-///
-/// By requiring a visitor we ensure that the caller has to handle both
-/// scenarios, even if one involves erroring. A type like
-/// [Cow][std::borrow::Cow] is an example of a type which can comfortably handle
-/// both.
-pub trait ValueVisitor<'de>: Sized {
-    /// The value being visited.
-    type Target: ?Sized + ToOwned;
-    /// The value produced.
-    type Ok;
-    /// The error produced.
+/// Trait that allows a type to be repeatedly coerced into a decoder.
+pub trait AsDecoder {
+    /// Error type raised by calling `as_decoder`.
     type Error: Error;
+    /// The decoder we reborrow as.
+    type Decoder<'this>: Decoder<'this, Error = Self::Error>
+    where
+        Self: 'this;
 
-    /// Format an error indicating what was expected by this visitor.
-    ///
-    /// Override to be more specific about the type that failed.
-    fn expecting(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result;
-
-    /// Visit an owned value.
-    #[inline]
-    fn visit_owned(self, value: <Self::Target as ToOwned>::Owned) -> Result<Self::Ok, Self::Error> {
-        self.visit_any(value.borrow())
-    }
-
-    /// Visit a string that is borrowed directly from the source data.
-    #[inline]
-    fn visit_borrowed(self, value: &'de Self::Target) -> Result<Self::Ok, Self::Error> {
-        self.visit_any(value)
-    }
-
-    /// Visit a string that is provided from the decoder in any manner possible.
-    /// Which might require additional decoding work.
-    #[inline]
-    fn visit_any(self, _: &Self::Target) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(BadVisitorType::new(
-            expecting::AnyValue,
-            &ReferenceVisistorExpecting(self),
-        )))
-    }
+    /// Borrow self as a new decoder.
+    fn as_decoder(&self) -> Result<Self::Decoder<'_>, Self::Error>;
 }
 
 /// A pack that can construct encoders.
@@ -66,6 +31,11 @@ pub trait PackDecoder<'de> {
     /// Return decoder to unpack the next element.
     #[must_use = "decoders must be consumed"]
     fn next(&mut self) -> Result<Self::Decoder<'_>, Self::Error>;
+
+    /// Stop decoding the current pack.
+    ///
+    /// This is required to call after a pack has finished decoding.
+    fn end(self) -> Result<(), Self::Error>;
 }
 
 /// Trait governing how to decode a sequence.
@@ -84,6 +54,11 @@ pub trait SequenceDecoder<'de> {
     /// Decode the next element.
     #[must_use = "decoders must be consumed"]
     fn next(&mut self) -> Result<Option<Self::Decoder<'_>>, Self::Error>;
+
+    /// Stop decoding the current sequence.
+    ///
+    /// This is required to call after a sequence has finished decoding.
+    fn end(self) -> Result<(), Self::Error>;
 }
 
 /// Trait governing how to decode a sequence of pairs.
@@ -102,9 +77,16 @@ pub trait PairsDecoder<'de> {
     /// Get a size hint of known remaining elements.
     fn size_hint(&self) -> Option<usize>;
 
-    /// Decode the next key. This returns `Ok(None)` where there are no more elements to decode.
+    /// Decode the next key. This returns `Ok(None)` where there are no more
+    /// elements to decode.
     #[must_use = "decoders must be consumed"]
     fn next(&mut self) -> Result<Option<Self::Decoder<'_>>, Self::Error>;
+
+    /// End the pair decoder.
+    ///
+    /// If there are any remaining elements in the sequence of pairs, this
+    /// indicates that they should be flushed.
+    fn end(self) -> Result<(), Self::Error>;
 }
 
 /// Trait governing how to decode a field.
@@ -176,6 +158,10 @@ pub trait VariantDecoder<'de> {
 pub trait Decoder<'de>: Sized {
     /// Error type raised by the decoder.
     type Error: Error;
+    /// The type returned when the decoder is buffered.
+    type Buffer: AsDecoder<Error = Self::Error>;
+    /// Decoder for a value that is present.
+    type Some: Decoder<'de, Error = Self::Error>;
     /// Pack decoder implementation.
     type Pack: PackDecoder<'de, Error = Self::Error>;
     /// Sequence decoder implementation.
@@ -184,18 +170,11 @@ pub trait Decoder<'de>: Sized {
     type Tuple: PackDecoder<'de, Error = Self::Error>;
     /// Map decoder implementation.
     type Map: PairsDecoder<'de, Error = Self::Error>;
-    /// Decoder for a value that is present.
-    type Some: Decoder<'de, Error = Self::Error>;
     /// Decoder for a struct.
     ///
     /// The caller receives a [PairsDecoder] which when advanced with
     /// [PairsDecoder::next] indicates the fields of the structure.
     type Struct: PairsDecoder<'de, Error = Self::Error>;
-    /// Decoder for a tuple struct.
-    ///
-    /// The caller receives a [PairsDecoder] which when advanced with
-    /// [PairsDecoder::next] indicates the elements in the tuple.
-    type TupleStruct: PairsDecoder<'de, Error = Self::Error>;
     /// Decoder for a variant.
     ///
     /// The caller receives a [PairDecoder] which when advanced with
@@ -218,14 +197,14 @@ pub trait Decoder<'de>: Sized {
     ///
     /// impl Decoder<'_> for MyDecoder {
     ///     type Error = String;
-    ///     type Pack = Never<Self>;
-    ///     type Sequence = Never<Self>;
-    ///     type Tuple = Never<Self>;
-    ///     type Map = Never<Self>;
-    ///     type Some = Never<Self>;
-    ///     type Struct = Never<Self>;
-    ///     type TupleStruct = Never<Self>;
-    ///     type Variant = Never<Self>;
+    ///     type Buffer = Never<Self::Error>;
+    ///     type Pack = Never<Self::Error>;
+    ///     type Sequence = Never<Self::Error>;
+    ///     type Tuple = Never<Self::Error>;
+    ///     type Map = Never<Self::Error>;
+    ///     type Some = Never<Self::Error>;
+    ///     type Struct = Never<Self::Error>;
+    ///     type Variant = Never<Self::Error>;
     ///
     ///     fn expecting(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
     ///         write!(f, "32-bit unsigned integers")
@@ -238,6 +217,89 @@ pub trait Decoder<'de>: Sized {
     /// ```
     fn expecting(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result;
 
+    /// Return a [TypeHint] indicating which type is being produced by the
+    /// [Decoder].
+    ///
+    /// Not all formats support type hints, and they might be ranging from
+    /// detailed (`a 32-bit unsigned integer`) to vague (`a number`).
+    ///
+    /// This is used to construct dynamic containers of types.
+    fn type_hint(&mut self) -> Result<TypeHint, Self::Error> {
+        Ok(TypeHint::Any)
+    }
+
+    /// Buffer the current decoder into a buffer that can be used multiple times.
+    ///
+    /// Buffering a decoder is necessary when additional introspection is needed
+    /// to decode a type, but it also means that:
+    ///
+    /// * The entire contents of the decoder needs to be dynamically buffered in
+    ///   memory.
+    /// * The in-memory representation might be lossy in some trivial ways. Such
+    ///   as arbitrary precision numbers being punted into a 64-bit float.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use musli::{Decode, Decoder, Mode};
+    /// use musli::de::{AsDecoder, PairsDecoder, PairDecoder};
+    /// use musli::error::Error;
+    ///
+    /// #[derive(Decode)]
+    /// struct Variant2 {
+    ///     name: String,
+    ///     age: u32,
+    /// }
+    ///
+    /// enum MyVariantType {
+    ///     Variant1,
+    ///     Variant2(Variant2),
+    /// }
+    ///
+    /// impl<'de, M> Decode<'de, M> for MyVariantType where M: Mode {
+    ///     fn decode<D>(decoder: D) -> Result<Self, D::Error>
+    ///     where
+    ///         D: Decoder<'de>,
+    ///     {
+    ///         let mut buffer = decoder.decode_buffer::<M>()?;
+    ///
+    ///         let mut st = buffer.as_decoder()?.decode_map()?;
+    ///
+    ///         let mut discriminator = None::<u32>;
+    ///
+    ///         while let Some(mut e) = st.next()? {
+    ///             let found = e.first()?.decode_string(musli::utils::visit_string_fn(|string| {
+    ///                 Ok(string == "type")
+    ///             }))?;
+    ///
+    ///             if found {
+    ///                 discriminator = Some(e.second().and_then(Decode::<M>::decode)?);
+    ///                 break;
+    ///             }
+    ///         }
+    ///
+    ///         st.end()?;
+    ///
+    ///         match discriminator {
+    ///             Some(0) => Ok(MyVariantType::Variant1),
+    ///             Some(1) => Ok(MyVariantType::Variant2(buffer.as_decoder().and_then(Decode::<M>::decode)?)),
+    ///             Some(other) => Err(D::Error::invalid_variant_tag("MyVariantType", other)),
+    ///             None => Err(D::Error::missing_variant_tag("MyVariantType")),
+    ///         }
+    ///     }
+    /// }
+    /// ```
+    #[inline]
+    fn decode_buffer<M>(self) -> Result<Self::Buffer, Self::Error>
+    where
+        M: Mode,
+    {
+        Err(Self::Error::message(format_args!(
+            "buffering not supported, expected {}",
+            ExpectingWrapper(self).format()
+        )))
+    }
+
     /// Decode a unit or something that is empty.
     ///
     /// # Examples
@@ -259,8 +321,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_unit(self) -> Result<(), Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Unit,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Unit,
             &ExpectingWrapper(self),
         )))
     }
@@ -289,8 +351,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_bool(self) -> Result<bool, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Bool,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Bool,
             &ExpectingWrapper(self),
         )))
     }
@@ -319,8 +381,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_char(self) -> Result<char, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Char,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Char,
             &ExpectingWrapper(self),
         )))
     }
@@ -349,8 +411,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_u8(self) -> Result<u8, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Unsigned8,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Unsigned8,
             &ExpectingWrapper(self),
         )))
     }
@@ -379,8 +441,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_u16(self) -> Result<u16, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Unsigned16,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Unsigned16,
             &ExpectingWrapper(self),
         )))
     }
@@ -409,8 +471,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_u32(self) -> Result<u32, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Unsigned32,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Unsigned32,
             &ExpectingWrapper(self),
         )))
     }
@@ -439,8 +501,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_u64(self) -> Result<u64, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Unsigned64,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Unsigned64,
             &ExpectingWrapper(self),
         )))
     }
@@ -469,8 +531,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_u128(self) -> Result<u128, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Unsigned128,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Unsigned128,
             &ExpectingWrapper(self),
         )))
     }
@@ -499,8 +561,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_i8(self) -> Result<i8, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Signed8,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Signed8,
             &ExpectingWrapper(self),
         )))
     }
@@ -529,8 +591,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_i16(self) -> Result<i16, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Signed16,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Signed16,
             &ExpectingWrapper(self),
         )))
     }
@@ -559,8 +621,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_i32(self) -> Result<i32, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Signed32,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Signed32,
             &ExpectingWrapper(self),
         )))
     }
@@ -589,8 +651,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_i64(self) -> Result<i64, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Signed64,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Signed64,
             &ExpectingWrapper(self),
         )))
     }
@@ -619,8 +681,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_i128(self) -> Result<i128, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Signed128,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Signed128,
             &ExpectingWrapper(self),
         )))
     }
@@ -649,8 +711,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_usize(self) -> Result<usize, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Usize,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Usize,
             &ExpectingWrapper(self),
         )))
     }
@@ -679,8 +741,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_isize(self) -> Result<isize, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Isize,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Isize,
             &ExpectingWrapper(self),
         )))
     }
@@ -709,8 +771,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_f32(self) -> Result<f32, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Float32,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Float32,
             &ExpectingWrapper(self),
         )))
     }
@@ -739,8 +801,21 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_f64(self) -> Result<f64, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Float64,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Float64,
+            &ExpectingWrapper(self),
+        )))
+    }
+
+    /// Decode an unknown number using a visitor that can handle arbitrary
+    /// precision numbers.
+    #[inline]
+    fn decode_number<V>(self, _: V) -> Result<V::Ok, V::Error>
+    where
+        V: NumberVisitor<Error = Self::Error>,
+    {
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Number,
             &ExpectingWrapper(self),
         )))
     }
@@ -769,8 +844,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_array<const N: usize>(self) -> Result<[u8; N], Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Array,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Array,
             &ExpectingWrapper(self),
         )))
     }
@@ -829,8 +904,8 @@ pub trait Decoder<'de>: Sized {
     where
         V: ValueVisitor<'de, Target = [u8], Error = Self::Error>,
     {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Bytes,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Bytes,
             &ExpectingWrapper(self),
         )))
     }
@@ -889,8 +964,8 @@ pub trait Decoder<'de>: Sized {
     where
         V: ValueVisitor<'de, Target = str, Error = Self::Error>,
     {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::String,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::String,
             &ExpectingWrapper(self),
         )))
     }
@@ -926,8 +1001,8 @@ pub trait Decoder<'de>: Sized {
     #[inline]
     #[must_use = "decoders must be consumed"]
     fn decode_option(self) -> Result<Option<Self::Some>, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Option,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Option,
             &ExpectingWrapper(self),
         )))
     }
@@ -957,6 +1032,7 @@ pub trait Decoder<'de>: Sized {
     ///         let mut unpack = decoder.decode_pack()?;
     ///         let field = unpack.next().and_then(Decode::<M>::decode)?;
     ///         let data = unpack.next().and_then(Decode::<M>::decode)?;
+    ///         unpack.end()?;
     ///
     ///         Ok(Self {
     ///             field,
@@ -967,8 +1043,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_pack(self) -> Result<Self::Pack, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Pack,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Pack,
             &ExpectingWrapper(self),
         )))
     }
@@ -997,6 +1073,8 @@ pub trait Decoder<'de>: Sized {
     ///             data.push(<String as Decode<M>>::decode(decoder)?);
     ///         }
     ///
+    ///         seq.end()?;
+    ///
     ///         Ok(Self {
     ///             data
     ///         })
@@ -1005,15 +1083,13 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_sequence(self) -> Result<Self::Sequence, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Sequence,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Sequence,
             &ExpectingWrapper(self),
         )))
     }
 
-    /// Return a helper to decode a tuple.
-    ///
-    /// A tuple is a fixed-length sequence.
+    /// Decode a fixed-length sequence of elements of length `len`.
     ///
     /// # Examples
     ///
@@ -1034,19 +1110,23 @@ pub trait Decoder<'de>: Sized {
     ///         let mut tuple = decoder.decode_tuple(2)?;
     ///         let string = tuple.next().and_then(<String as Decode<M>>::decode)?;
     ///         let integer = tuple.next().and_then(<u32 as Decode<M>>::decode)?;
+    ///         tuple.end()?;
     ///         Ok(Self(string, integer))
     ///     }
     /// }
     /// ```
     #[inline]
-    fn decode_tuple(self, _: usize) -> Result<Self::Tuple, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Tuple,
+    fn decode_tuple(self, #[allow(unused)] len: usize) -> Result<Self::Tuple, Self::Error> {
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Tuple,
             &ExpectingWrapper(self),
         )))
     }
 
-    /// Decode a map.
+    /// Decode a map of unknown length.
+    ///
+    /// The length of the map must somehow be determined from the underlying
+    /// format.
     ///
     /// # Examples
     ///
@@ -1074,6 +1154,8 @@ pub trait Decoder<'de>: Sized {
     ///             data.insert(key, value);
     ///         }
     ///
+    ///         map.end()?;
+    ///
     ///         Ok(Self {
     ///             data
     ///         })
@@ -1082,13 +1164,19 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_map(self) -> Result<Self::Map, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Map,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Map,
             &ExpectingWrapper(self),
         )))
     }
 
-    /// Return a helper to decode a struct with named fields.
+    /// Decode a struct which has an expected `len` number of elements.
+    ///
+    /// The `len` indicates how many fields the decoder is *expecting* depending
+    /// on how many fields are present in the underlying struct being decoded,
+    /// butit should only be considered advisory.
+    ///
+    /// The size of a struct might therefore change from one session to another.
     ///
     /// # Examples
     ///
@@ -1109,7 +1197,7 @@ pub trait Decoder<'de>: Sized {
     ///     where
     ///         D: Decoder<'de>,
     ///     {
-    ///         let mut st = decoder.decode_struct(2)?;
+    ///         let mut st = decoder.decode_map()?;
     ///         let mut string = None;
     ///         let mut integer = None;
     ///
@@ -1130,6 +1218,8 @@ pub trait Decoder<'de>: Sized {
     ///             }
     ///         }
     ///
+    ///         st.end()?;
+    ///
     ///         Ok(Self {
     ///             string: string.ok_or_else(|| D::Error::expected_tag("Struct", "string"))?,
     ///             integer: integer.ok_or_else(|| D::Error::expected_tag("Struct", "integer"))?,
@@ -1139,88 +1229,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_struct(self, _: usize) -> Result<Self::Struct, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Struct,
-            &ExpectingWrapper(self),
-        )))
-    }
-
-    /// Return a helper to decode a tuple struct.
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// use std::collections::HashMap;
-    ///
-    /// use musli::de::{Decode, Decoder, PairsDecoder, PairDecoder};
-    /// use musli::error::Error;
-    /// use musli::mode::Mode;
-    ///
-    /// struct TupleStruct(String, u32);
-    ///
-    /// impl<'de, M> Decode<'de, M> for TupleStruct where M: Mode {
-    ///     fn decode<D>(decoder: D) -> Result<Self, D::Error>
-    ///     where
-    ///         D: Decoder<'de>,
-    ///     {
-    ///         let mut st = decoder.decode_tuple_struct(2)?;
-    ///         let mut string = None;
-    ///         let mut integer = None;
-    ///
-    ///         while let Some(mut entry) = st.next()? {
-    ///             let tag = entry.first().and_then(<usize as Decode<M>>::decode)?;
-    ///
-    ///             match tag {
-    ///                 0 => {
-    ///                     string = Some(entry.second().and_then(<String as Decode<M>>::decode)?);
-    ///                 }
-    ///                 1 => {
-    ///                     integer = Some(entry.second().and_then(<u32 as Decode<M>>::decode)?);
-    ///                 }
-    ///                 tag => {
-    ///                     return Err(D::Error::invalid_field_tag("Struct", tag))
-    ///                 }
-    ///             }
-    ///         }
-    ///
-    ///         let string = string.ok_or_else(|| D::Error::expected_tag("Struct", "string"))?;
-    ///         let integer = integer.ok_or_else(|| D::Error::expected_tag("Struct", "integer"))?;
-    ///
-    ///         Ok(Self(string, integer))
-    ///     }
-    /// }
-    /// ```
-    #[inline]
-    fn decode_tuple_struct(self, _: usize) -> Result<Self::TupleStruct, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::TupleStruct,
-            &ExpectingWrapper(self),
-        )))
-    }
-
-    /// Decode a unit variant.
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// use musli::{Decode, Decoder, Mode};
-    ///
-    /// struct UnitStruct;
-    ///
-    /// impl<'de, M> Decode<'de, M> for UnitStruct where M: Mode {
-    ///     fn decode<D>(decoder: D) -> Result<Self, D::Error>
-    ///     where
-    ///         D: Decoder<'de>,
-    ///     {
-    ///         decoder.decode_unit_struct()?;
-    ///         Ok(Self)
-    ///     }
-    /// }
-    /// ```
-    #[inline]
-    fn decode_unit_struct(self) -> Result<(), Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::UnitStruct,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Struct,
             &ExpectingWrapper(self),
         )))
     }
@@ -1266,8 +1276,8 @@ pub trait Decoder<'de>: Sized {
     /// ```
     #[inline]
     fn decode_variant(self) -> Result<Self::Variant, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Variant,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Variant,
             &ExpectingWrapper(self),
         )))
     }
@@ -1284,15 +1294,3 @@ where
         self.0.expecting(f)
     }
 }
-
-#[repr(transparent)]
-struct ReferenceVisistorExpecting<T>(T);
-
-impl<'de, T> Expecting for ReferenceVisistorExpecting<T>
-where
-    T: ValueVisitor<'de>,
-{
-    fn expecting(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        self.0.expecting(f)
-    }
-}
