@@ -2,7 +2,7 @@ use core::fmt;
 
 use crate::en::Encode;
 use crate::error::Error;
-use crate::expecting::{self, Expecting, InvalidType};
+use crate::expecting::{self, Expecting};
 use crate::mode::Mode;
 
 /// Trait governing how to encode a sequence.
@@ -81,7 +81,9 @@ pub trait PairEncoder {
         Self: 'this;
 
     /// The encoder returned when advancing the map encoder to encode the value.
-    type Second: Encoder<Ok = Self::Ok, Error = Self::Error>;
+    type Second<'this>: Encoder<Ok = Self::Ok, Error = Self::Error>
+    where
+        Self: 'this;
 
     /// Insert the pair immediately.
     #[inline]
@@ -93,7 +95,8 @@ pub trait PairEncoder {
         S: Encode<M>,
     {
         self.first().and_then(|e| first.encode(e))?;
-        self.second().and_then(|e| second.encode(e))
+        self.second().and_then(|e| second.encode(e))?;
+        self.end()
     }
 
     /// Return the encoder for the first element in the pair.
@@ -102,7 +105,10 @@ pub trait PairEncoder {
 
     /// Return encoder for the second element in the pair.
     #[must_use = "encoders must be consumed"]
-    fn second(self) -> Result<Self::Second, Self::Error>;
+    fn second(&mut self) -> Result<Self::Second<'_>, Self::Error>;
+
+    /// Stop encoding this pair.
+    fn end(self) -> Result<Self::Ok, Self::Error>;
 }
 
 /// Trait governing how to encode a variant.
@@ -156,10 +162,10 @@ pub trait Encoder: Sized {
     type Ok;
     /// The error raised by an encoder.
     type Error: Error;
-    /// A simple pack that packs a sequence of elements.
-    type Pack: SequenceEncoder<Ok = Self::Ok, Error = Self::Error>;
     /// Encoder returned when encoding an optional value which is present.
     type Some: Encoder<Ok = Self::Ok, Error = Self::Error>;
+    /// A simple pack that packs a sequence of elements.
+    type Pack: SequenceEncoder<Ok = Self::Ok, Error = Self::Error>;
     /// The type of a sequence encoder.
     type Sequence: SequenceEncoder<Ok = Self::Ok, Error = Self::Error>;
     /// The type of a tuple encoder.
@@ -168,8 +174,6 @@ pub trait Encoder: Sized {
     type Map: PairsEncoder<Ok = Self::Ok, Error = Self::Error>;
     /// Encoder that can encode a struct.
     type Struct: PairsEncoder<Ok = Self::Ok, Error = Self::Error>;
-    /// Encoder that can encode a tuple struct.
-    type TupleStruct: PairsEncoder<Ok = Self::Ok, Error = Self::Error>;
     /// Encoder for a struct variant.
     type Variant: VariantEncoder<Ok = Self::Ok, Error = Self::Error>;
 
@@ -197,8 +201,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_unit(self) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Unit,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Unit,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -225,8 +229,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_bool(self, _: bool) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Bool,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Bool,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -253,8 +257,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_char(self, _: char) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Char,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Char,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -281,8 +285,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_u8(self, _: u8) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Unsigned8,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Unsigned8,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -309,8 +313,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_u16(self, _: u16) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Unsigned16,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Unsigned16,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -337,8 +341,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_u32(self, _: u32) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Unsigned32,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Unsigned32,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -365,8 +369,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_u64(self, _: u64) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Unsigned64,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Unsigned64,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -393,8 +397,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_u128(self, _: u128) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Unsigned128,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Unsigned128,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -421,8 +425,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_i8(self, _: i8) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Signed8,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Signed8,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -449,8 +453,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_i16(self, _: i16) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Signed16,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Signed16,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -477,8 +481,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_i32(self, _: i32) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Signed32,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Signed32,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -505,8 +509,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_i64(self, _: i64) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Signed64,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Signed64,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -533,8 +537,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_i128(self, _: i128) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Signed128,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Signed128,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -561,8 +565,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_usize(self, _: usize) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Usize,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Usize,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -589,8 +593,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_isize(self, _: isize) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Isize,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Isize,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -617,8 +621,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_f32(self, _: f32) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Float32,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Float32,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -645,8 +649,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_f64(self, _: f64) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Float64,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Float64,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -673,8 +677,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_array<const N: usize>(self, _: [u8; N]) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Array,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Array,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -701,8 +705,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_bytes(self, _: &[u8]) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Bytes,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Bytes,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -737,8 +741,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_bytes_vectored(self, _: &[&[u8]]) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Bytes,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Bytes,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -765,8 +769,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_string(self, _: &str) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::String,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::String,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -800,8 +804,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_some(self) -> Result<Self::Some, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Option,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Option,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -835,8 +839,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_none(self) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Option,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Option,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -873,13 +877,19 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_pack(self) -> Result<Self::Pack, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Pack,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Pack,
             &ExpectingWrapper::new(self),
         )))
     }
 
-    /// Encode a sequence with a known length.
+    /// Encode a sequence with a known length `len`.
+    ///
+    /// A sequence encodes one element following another and must in some way
+    /// encode the length of the sequence in the underlying format. It is
+    /// decoded with [Decoder::decode_sequence].
+    ///
+    /// [Decoder::decode_sequence]: crate::de::Decoder::decode_sequence
     ///
     /// # Examples
     ///
@@ -907,14 +917,22 @@ pub trait Encoder: Sized {
     /// }
     /// ```
     #[inline]
-    fn encode_sequence(self, _: usize) -> Result<Self::Sequence, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Sequence,
+    fn encode_sequence(self, #[allow(unused)] len: usize) -> Result<Self::Sequence, Self::Error> {
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Sequence,
             &ExpectingWrapper::new(self),
         )))
     }
 
-    /// Encode a tuple.
+    /// Encode a tuple with a known length `len`.
+    ///
+    /// This is almost identical to [Encoder::encode_sequence] except that we
+    /// know that we are encoding a fixed-length container of length `len`, and
+    /// assuming the size of that container doesn't change in size it can be
+    /// decoded using [Decoder::decode_tuple] again without the underlying
+    /// format having to encode the size of the container.
+    ///
+    /// [Decoder::decode_tuple]: crate::de::Decoder::decode_tuple
     ///
     /// # Examples
     ///
@@ -937,18 +955,20 @@ pub trait Encoder: Sized {
     /// }
     /// ```
     #[inline]
-    fn encode_tuple(self, _: usize) -> Result<Self::Tuple, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Tuple,
+    fn encode_tuple(self, #[allow(unused)] len: usize) -> Result<Self::Tuple, Self::Error> {
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Tuple,
             &ExpectingWrapper::new(self),
         )))
     }
 
-    /// Encode a map with a known length.
+    /// Encode a map with a known length `len`.
+    ///
+    ///
     #[inline]
-    fn encode_map(self, _: usize) -> Result<Self::Map, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Map,
+    fn encode_map(self, #[allow(unused)] len: usize) -> Result<Self::Map, Self::Error> {
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Map,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -961,12 +981,12 @@ pub trait Encoder: Sized {
     /// use musli::en::{Encode, Encoder, PairEncoder, PairsEncoder};
     /// use musli::mode::Mode;
     ///
-    /// struct TupleStruct {
+    /// struct Struct {
     ///     name: String,
     ///     age: u32,
     /// }
     ///
-    /// impl<M> Encode<M> for TupleStruct where M: Mode {
+    /// impl<M> Encode<M> for Struct where M: Mode {
     ///     fn encode<E>(&self, encoder: E) -> Result<E::Ok, E::Error>
     ///     where
     ///         E: Encoder
@@ -980,63 +1000,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_struct(self, _: usize) -> Result<Self::Struct, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Struct,
-            &ExpectingWrapper::new(self),
-        )))
-    }
-
-    /// Encode a tuple struct.
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// use musli::en::{Encode, Encoder, PairEncoder, PairsEncoder};
-    /// use musli::mode::Mode;
-    ///
-    /// struct TupleStruct(String);
-    ///
-    /// impl<M> Encode<M> for TupleStruct where M: Mode {
-    ///     fn encode<E>(&self, encoder: E) -> Result<E::Ok, E::Error>
-    ///     where
-    ///         E: Encoder
-    ///     {
-    ///         let mut tuple = encoder.encode_tuple_struct(1)?;
-    ///         tuple.insert::<M, _, _>(0usize, &self.0)?;
-    ///         tuple.end()
-    ///     }
-    /// }
-    /// ```
-    #[inline]
-    fn encode_tuple_struct(self, _: usize) -> Result<Self::TupleStruct, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::TupleStruct,
-            &ExpectingWrapper::new(self),
-        )))
-    }
-
-    /// Encode a unit struct.
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// use musli::{Encode, Encoder, Mode};
-    ///
-    /// struct UnitStruct;
-    ///
-    /// impl<M> Encode<M> for UnitStruct where M: Mode {
-    ///     fn encode<E>(&self, encoder: E) -> Result<E::Ok, E::Error>
-    ///     where
-    ///         E: Encoder
-    ///     {
-    ///         encoder.encode_unit_struct()
-    ///     }
-    /// }
-    /// ```
-    #[inline]
-    fn encode_unit_struct(self) -> Result<Self::Ok, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::UnitStruct,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Struct,
             &ExpectingWrapper::new(self),
         )))
     }
@@ -1088,8 +1053,8 @@ pub trait Encoder: Sized {
     /// ```
     #[inline]
     fn encode_variant(self) -> Result<Self::Variant, Self::Error> {
-        Err(Self::Error::message(InvalidType::new(
-            expecting::Variant,
+        Err(Self::Error::message(expecting::invalid_type(
+            &expecting::Variant,
             &ExpectingWrapper::new(self),
         )))
     }
